
local modname = (...)
local m = {}
package.loaded[modname] = m

-- TODO:
--   should use some logging mechanism instead of `print'.
--   also a logging library allows verbosity filter.

local throw = error

-- forward declaration, shared by public and private codes.
local groupregistry, groupstack
local currentgroup, currenttest
local dotest
-- unique objects as throw/catch filter
--local assertionfailed = newproxy()
local assertionfailed = {}

do -- public api

   function m.group (name)
      if name ~= nil then
         currentgroup = groupregistry[name]
         groupstack[#groupstack] = currentgroup
      end
      return currentgroup
   end

   function m.grouppush (name)
      name = name or "default"
      currentgroup = groupregistry[name]
      groupstack[#groupstack + 1] = currentgroup
   end

   function m.grouppop ()
      groupstack[#groupstack] = nil
      currentgroup = groupstack[#groupstack]
   end

   function m.setup (f)
      currentgroup.setup = f
   end

   function m.teardown (f)
      currentgroup.teardown = f
   end

   function m.test (name, body)
      return currentgroup:addtest(name, body)
   end

   -- TODO: add traceback information using debug library
   function m.failbecause (reason, ...)
      local message = reason and reason:format(...) or "explicitly fail!"
      if currenttest ~= nil then
         assertionfailed.message = message
         throw(assertionfailed)
      end
   end

   function m.assert (value, predicate)
      if currenttest then
         currenttest.assertions = currenttest.assertions + 1
         local success, description = predicate(value)
         if success then
            return
         end
         return m.failbecause(description or "assertion failed")
      end
   end

   function m.rungroup (group)
      if group.status ~= "ready" then
         return
      end
      if group.numtests == 0 then
         group:ignore("empty group")
         print(">>>> [ignored] empty group <<" .. group.name .. ">>")
         return
      end
      print(">>>> [running] group <<" .. group.name .. ">>...")
      local setup, teardown = group.setup, group.teardown
      for _name, test in group:alltests() do
         if test.status == "ready" then
            dotest(test, setup, teardown)
         end
      end
      local stat = group:statistic()
      if stat.run > 0 then
         if stat.run == stat.passed then
            group.status = "passed"
         elseif stat.error > 0 then
            group.status = "error"
         else
            group.status = "failed"
         end
      else
         group.status = "passed"
      end
      print(">>>> [" .. group.status .. "] group <<" .. group.name .. ">>.")
   end

   function m.allgroups ()
      return ipairs(groupregistry)
   end

   function m.runall ()
      for _i, g in m.allgroups() do
         m.rungroup(g)
      end
   end

   function m.statistic ()
      local stat = {
         grouppassed = 0, groupfailed = 0, grouperror = 0,
         groupready =0, groupignored = 0, groupdisabled = 0,
         passed = 0, failed = 0, error = 0,
         ready = 0, run = 0, ignored = 0, disabled = 0,
      }
      for _, group in m.allgroups() do
         local groupstatus = "group" .. group.status
         stat[groupstatus] = stat[groupstatus] + 1
         for k, v in pairs(group:statistic()) do
            stat[k] = stat[k] + v
         end
      end
      stat.grouprun = stat.grouppassed + stat.groupfailed + stat.grouperror
      return stat
   end

end -- public api


do -- implementation

   -- declared on top and called by public API, no `local' here
   function dotest (test, setup, teardown)
      currenttest = test
      local success, thrown = true, "no error"
      if setup then
         success, thrown = pcall(setup)
      end
      if not success then
         test.status = "error"
         test.detail = thrown or "error in set up routine"
      else 
         success, thrown = pcall(test.body)
         if success then
            test.status = "passed"
         elseif thrown == assertionfailed then
            test.status = "failed"
            test.detail = assertionfailed.message
         else
            test.status = "error"
            test.detail = thrown or "unkown error"
         end
      end
      if teardown then
         pcall(teardown)
      end
      currenttest = nil
   end

   local function ignore (obj, reason)
      obj.status = "ignored"
      obj.detail = reason or "ignored"
   end

   local function disable (obj, reason)
      obj.status = "disabled"
      obj.detail = reason or "ignored"
   end

   local function newtest (name, body)
      return {
         name = name,
         body = body,
         assertions = 0,
         status = "ready",
         ignore = ignore,
         disable = disable,
      }
   end

   local function newgroup (name)
      local group = {
         name = name,
         tests = {},
         numtests = 0,
         status = "ready",
         ignore = ignore,
         disable = disable,
      }
      function group:addtest (name, body)
         local test = newtest(name, body)
         -- the order of tests is NOT retained on purpose.
         test.group = self
         self.tests[name] = test
         self.numtests = self.numtests + 1
         return test
      end
      function group:alltests ()
         return pairs(self.tests)
      end
      function group:statistic ()
         local stat = {
            passed = 0, failed = 0, error = 0,
            ready = 0, ignored = 0, disabled = 0,
         }
         for _name, test in self:alltests() do
            local status = test.status
            stat[status] = stat[status] + 1
         end
         stat.run = stat.passed + stat.failed + stat.error
         return stat
      end
      return group
   end

   local function creategroupregistry ()
      groupstack = {}
      groupregistry = setmetatable ({}, {
         __index = function (registry, name)
            local group = newgroup (name)
            registry[#registry + 1] = group
            registry[name] = group
            return group
         end
      })
      -- create the default group
      m.grouppush "default"
   end

   local function createbackdoor ()
      local backdoor = {}
      backdoor.reinitialize = function () return creategroupregistry () end
      backdoor.currentgroup = function () return currentgroup end
      backdoor.currenttest = function () return currenttest end
      m.__backdoor = backdoor
   end

   creategroupregistry()
   createbackdoor()

end -- implementation

